@@ -2,9 +2,8 @@ import datetime
 import decimal
 import enum
 import pathlib
-import re
-import sys
 import uuid
+import warnings
 from abc import ABCMeta
 from collections.abc import Callable as abcCallable
 from types import FunctionType
@@ -29,7 +28,7 @@ from typing import (
 
 import bson
 import pydantic
-from pydantic.error_wrappers import ValidationError
+from pydantic.error_wrappers import ErrorWrapper, ValidationError
 from pydantic.fields import Field as PDField
 from pydantic.fields import FieldInfo as PDFieldInfo
 from pydantic.fields import Undefined
@@ -39,11 +38,17 @@ from pydantic.utils import lenient_issubclass
 
 from odmantic.bson import (
     _BSON_SUBSTITUTED_FIELDS,
-    BSON_TYPES_ENCODERS,
     BaseBSONModel,
     ObjectId,
     _decimalDecimal,
 )
+from odmantic.config import BaseODMConfig, validate_config
+from odmantic.exceptions import (
+    DocumentParsingError,
+    ErrorList,
+    KeyNotFoundInDocumentError,
+    ReferencedDocumentNotFoundError,
+)
 from odmantic.field import (
     FieldProxy,
     ODMBaseField,
@@ -53,6 +58,13 @@ from odmantic.field import (
     ODMReference,
 )
 from odmantic.reference import ODMReferenceInfo
+from odmantic.typing import USES_OLD_TYPING_INTERFACE
+from odmantic.utils import (
+    is_dunder,
+    raise_on_invalid_collection_name,
+    raise_on_invalid_key_name,
+    to_snake_case,
+)
 
 if TYPE_CHECKING:
 
@@ -63,12 +75,11 @@ if TYPE_CHECKING:
         ReprArgs,
     )
 
-USES_OLD_TYPING_INTERFACE = sys.version_info[:3] < (3, 7, 0)  # PEP 560
 if USES_OLD_TYPING_INTERFACE:
     from typing import _subs_tree  # type: ignore  # noqa
 
 
-UNTOUCHED_TYPES = FunctionType, property, classmethod, staticmethod
+UNTOUCHED_TYPES = FunctionType, property, classmethod, staticmethod, type
 
 
 def should_touch_field(value: Any = None, type_: Optional[Type] = None) -> bool:
@@ -79,23 +90,6 @@ def should_touch_field(value: Any = None, type_: Optional[Type] = None) -> bool:
     )
 
 
-def is_valid_odm_field_name(name: str) -> bool:
-    return not name.startswith("__") and not name.endswith("__")
-
-
-def raise_on_invalid_key_name(name: str) -> None:
-    # https://docs.mongodb.com/manual/reference/limits/#Restrictions-on-Field-Names
-    if name.startswith("$"):
-        raise TypeError("key_name cannot start with the dollar sign ($) character")
-    if "." in name:
-        raise TypeError("key_name cannot contain the dot (.) character")
-
-
-def to_snake_case(s: str) -> str:
-    tmp = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
-    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", tmp).lower()
-
-
 def find_duplicate_key(fields: Iterable[ODMBaseField]) -> Optional[str]:
     seen: Set[str] = set()
     for f in fields:
@@ -147,7 +141,11 @@ def is_type_mutable(type_: Type) -> bool:
         return not lenient_issubclass(type_origin, _IMMUTABLE_TYPES)
     else:
         return not (
-            type_ is None or lenient_issubclass(type_, _IMMUTABLE_TYPES)  # type:ignore
+            type_ is None  # type:ignore
+            or (
+                lenient_issubclass(type_, _IMMUTABLE_TYPES)
+                and not lenient_issubclass(type_, EmbeddedModel)
+            )
         )
 
 
@@ -202,6 +200,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
         annotations = resolve_annotations(
             namespace.get("__annotations__", {}), namespace.get("__module__")
         )
+        config = validate_config(namespace.get("Config", BaseODMConfig), name)
         odm_fields: Dict[str, ODMBaseField] = {}
         references: List[str] = []
         bson_serialized_fields: Set[str] = set()
@@ -211,7 +210,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
         for field_name, value in namespace.items():
             if (
                 should_touch_field(value=value)
-                and is_valid_odm_field_name(field_name)
+                and not is_dunder(field_name)
                 and field_name not in annotations
             ):
                 raise TypeError(
@@ -220,9 +219,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
 
         # Validate fields types and substitute bson fields
         for (field_name, field_type) in annotations.items():
-            if is_valid_odm_field_name(field_name) and should_touch_field(
-                type_=field_type
-            ):
+            if not is_dunder(field_name) and should_touch_field(type_=field_type):
                 substituted_type = validate_type(field_type)
                 # Handle BSON serialized fields after substitution to allow some
                 # builtin substitution
@@ -235,15 +232,16 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
         for (field_name, field_type) in annotations.items():
             value = namespace.get(field_name, Undefined)
 
-            if not is_valid_odm_field_name(field_name) or not should_touch_field(
-                value, field_type
-            ):
+            if is_dunder(field_name) or not should_touch_field(value, field_type):
                 continue  # pragma: no cover
                 # https://github.com/nedbat/coveragepy/issues/198
 
             if isinstance(value, PDFieldInfo):
                 raise TypeError("please use odmantic.Field instead of pydantic.Field")
 
+            if is_type_mutable(field_type):
+                mutable_fields.add(field_name)
+
             if lenient_issubclass(field_type, EmbeddedModel):
                 if isinstance(value, ODMFieldInfo):
                     namespace[field_name] = value.pydantic_field_info
@@ -256,7 +254,10 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
                     primary_field = False
 
                 odm_fields[field_name] = ODMEmbedded(
-                    primary_field=primary_field, model=field_type, key_name=key_name
+                    primary_field=primary_field,
+                    model=field_type,
+                    key_name=key_name,
+                    model_config=config,
                 )
             elif lenient_issubclass(field_type, Model):
                 if not isinstance(value, ODMReferenceInfo):
@@ -267,26 +268,26 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
                 key_name = value.key_name if value.key_name is not None else field_name
                 raise_on_invalid_key_name(key_name)
                 odm_fields[field_name] = ODMReference(
-                    model=field_type, key_name=key_name
+                    model=field_type, key_name=key_name, model_config=config
                 )
                 references.append(field_name)
                 del namespace[field_name]  # Remove default ODMReferenceInfo value
             else:
-                if is_type_mutable(field_type):
-                    mutable_fields.add(field_name)
                 if isinstance(value, ODMFieldInfo):
                     key_name = (
                         value.key_name if value.key_name is not None else field_name
                     )
                     raise_on_invalid_key_name(key_name)
                     odm_fields[field_name] = ODMField(
-                        primary_field=value.primary_field, key_name=key_name
+                        primary_field=value.primary_field,
+                        key_name=key_name,
+                        model_config=config,
                     )
                     namespace[field_name] = value.pydantic_field_info
 
                 elif value is Undefined:
                     odm_fields[field_name] = ODMField(
-                        primary_field=False, key_name=field_name
+                        primary_field=False, key_name=field_name, model_config=config
                     )
 
                 else:
@@ -298,7 +299,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
                             f" = {repr(value)}"
                         )
                     odm_fields[field_name] = ODMField(
-                        primary_field=False, key_name=field_name
+                        primary_field=False, key_name=field_name, model_config=config
                     )
 
         duplicate_key = find_duplicate_key(odm_fields.values())
@@ -311,6 +312,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
         namespace["__references__"] = tuple(references)
         namespace["__bson_serialized_fields__"] = frozenset(bson_serialized_fields)
         namespace["__mutable_fields__"] = frozenset(mutable_fields)
+        namespace["Config"] = config
 
     @no_type_check
     def __new__(
@@ -327,6 +329,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
             "Model",
             "EmbeddedModel",
         )
+
         if is_custom_cls:
             # Handle calls from pydantic.main.create_model (used internally by FastAPI)
             patched_bases = []
@@ -344,6 +347,7 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
         cls = super().__new__(mcs, name, bases, namespace, **kwargs)
 
         if is_custom_cls:
+            config: BaseODMConfig = namespace["Config"]
             # Patch Model related fields to build a "pure" pydantic model
             odm_fields: Dict[str, ODMBaseField] = namespace["__odm_fields__"]
             for field_name, field in odm_fields.items():
@@ -356,10 +360,12 @@ class BaseModelMetaclass(pydantic.main.ModelMetaclass):
                 mcs, f"{name}.__pydantic_model__", (BaseBSONModel,), namespace, **kwargs
             )
             # Change the title to generate clean JSON schemas from this "pure" model
-            pydantic_cls.__config__.title = name
-
+            if config.title is None:
+                pydantic_cls.__config__.title = name
             cls.__pydantic_model__ = pydantic_cls
+
             for name, field in cls.__odm_fields__.items():
+                field.bind_pydantic_field(cls.__fields__[name])
                 setattr(cls, name, FieldProxy(parent=None, field=field))
 
         return cls
@@ -378,6 +384,7 @@ class ModelMetaclass(BaseModelMetaclass):
             "__qualname__"
         ) not in ("_BaseODMModel", "Model"):
             mcs.__validate_cls_namespace__(name, namespace)
+            config: BaseODMConfig = namespace["Config"]
             primary_field: Optional[str] = None
             odm_fields: Dict[str, ODMBaseField] = namespace["__odm_fields__"]
 
@@ -393,35 +400,31 @@ class ModelMetaclass(BaseModelMetaclass):
                         "field already exists"
                     )
                 primary_field = "id"
-                odm_fields["id"] = ODMField(primary_field=True, key_name="_id")
+                odm_fields["id"] = ODMField(
+                    primary_field=True, key_name="_id", model_config=config
+                )
                 namespace["id"] = PDField(default_factory=ObjectId)
                 namespace["__annotations__"]["id"] = ObjectId
 
             namespace["__primary_field__"] = primary_field
 
-            if "__collection__" in namespace:
+            if config.collection is not None:
+                collection_name = config.collection
+            elif "__collection__" in namespace:
                 collection_name = namespace["__collection__"]
+                warnings.warn(
+                    "Defining the collection name with `__collection__` is deprecated. "
+                    "Please use `collection` config attribute instead.",
+                    DeprecationWarning,
+                )
             else:
                 cls_name = name
                 if cls_name.endswith("Model"):
                     # TODO document this
                     cls_name = cls_name[:-5]  # Strip Model in the class name
                 collection_name = to_snake_case(cls_name)
-                namespace["__collection__"] = collection_name
-
-            # Validate the collection name
-            # https://docs.mongodb.com/manual/reference/limits/#Restriction-on-Collection-Names
-            if "$" in collection_name:
-                raise TypeError(
-                    f"Invalid collection name for {name}: cannot contain '$'"
-                )
-            if collection_name == "":
-                raise TypeError(f"Invalid collection name for {name}: cannot be empty")
-            if collection_name.startswith("system."):
-                raise TypeError(
-                    f"Invalid collection name for {name}:"
-                    " cannot start with 'system.'"
-                )
+            raise_on_invalid_collection_name(collection_name, cls_name=name)
+            namespace["__collection__"] = collection_name
 
         return super().__new__(mcs, name, bases, namespace, **kwargs)
 
@@ -474,10 +477,9 @@ class _BaseODMModel(pydantic.BaseModel, metaclass=ABCMeta):
 
     __slots__ = ("__fields_modified__",)
 
-    class Config:
-        validate_all = True
-        validate_assignment = True
-        json_encoders = BSON_TYPES_ENCODERS
+    def __init__(self, **data: Any):
+        super().__init__(**data)
+        object.__setattr__(self, "__fields_modified__", set(self.__odm_fields__.keys()))
 
     @classmethod
     def validate(cls: Type[TBase], value: Any) -> TBase:
@@ -546,17 +548,69 @@ class _BaseODMModel(pydantic.BaseModel, metaclass=ABCMeta):
         Args:
             raw_doc: document to parse (as Dict)
 
+        Raises:
+            DocumentParsingError: the specified document is invalid
+
         Returns:
             an instance of the Model class this method is called on.
         """
-        doc: Dict[str, Any] = {}
+        errors, obj = cls._parse_doc_to_obj(raw_doc)
+        if len(errors) > 0:
+            raise DocumentParsingError(
+                errors=[errors],
+                model=cls,
+                primary_value=raw_doc.get("_id", "<unknown>"),
+            )
+        try:
+            instance = cls.parse_obj(obj)
+        except ValidationError as e:
+            raise DocumentParsingError(
+                errors=e.raw_errors,  # type: ignore
+                model=cls,
+                primary_value=raw_doc.get("_id", "<unknown>"),
+            )
+
+        return instance
+
+    @classmethod
+    def _parse_doc_to_obj(
+        cls: Type[TBase], raw_doc: Dict, base_loc: Tuple[str, ...] = ()
+    ) -> Tuple[ErrorList, Dict[str, Any]]:
+        errors: ErrorList = []
+        obj: Dict[str, Any] = {}
         for field_name, field in cls.__odm_fields__.items():
             if isinstance(field, ODMReference):
-                doc[field_name] = field.model.parse_doc(raw_doc[field.key_name])
+                sub_doc = raw_doc.get(field.key_name)
+                if sub_doc is None:
+                    errors.append(
+                        ErrorWrapper(
+                            exc=ReferencedDocumentNotFoundError(field.key_name),
+                            loc=base_loc + (field_name,),
+                        )
+                    )
+                else:
+                    sub_errors, sub_obj = field.model._parse_doc_to_obj(
+                        sub_doc, base_loc=base_loc + (field_name,)
+                    )
+                    errors.extend(sub_errors)
+                    obj[field_name] = sub_obj
             else:
-                doc[field_name] = raw_doc[field.key_name]
-        instance = cls.parse_obj(doc)
-        return instance
+                field = cast(Union[ODMField, ODMEmbedded], field)
+                value = raw_doc.get(field.key_name, Undefined)
+                if value is Undefined and not field.is_required_in_doc():
+                    value = field.get_default_importing_value()
+
+                if value is Undefined:
+                    errors.append(
+                        ErrorWrapper(
+                            exc=KeyNotFoundInDocumentError(field.key_name),
+                            loc=base_loc + (field_name,),
+                        )
+                    )
+                else:
+                    obj[field_name] = value
+
+        return errors, obj
 
 
 class Model(_BaseODMModel, metaclass=ModelMetaclass):
@@ -572,16 +626,6 @@ class Model(_BaseODMModel, metaclass=ModelMetaclass):
 
         id: Union[ObjectId, Any]  # TODO fix basic id field typing
 
-    def __init__(__odmantic_self__, **data: Any):
-        super().__init__(**data)
-        # Uses something other than `self` the first arg to allow "self" as a settable
-        # attribute
-        object.__setattr__(
-            __odmantic_self__,
-            "__fields_modified__",
-            set(__odmantic_self__.__odm_fields__.keys()),
-        )
-
     def __setattr__(self, name: str, value: Any) -> None:
         if name == self.__primary_field__:
             # TODO implement
